#ifndef CG_RENDERAREA_H
#define CG_RENDERAREA_H

#include <vector>
#include <list>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <iostream>

#include <QtOpenGL/QGLWidget>
#include <QtGui/QImage>
#include <QtCore/QTimer>

#include <GL/gl.h>
#include <GL/glu.h>

#include "mesh.h"

class GLwidget : public QGLWidget 
{
	static const GLdouble SV_TRANSZ;
	static const GLdouble PV_TRANSZ;

	// OBJ-File is translated to origin by (mesh.getBBoxMin() + mesh.getBBoxMax)/2, then scaled by scale and finally transflated to pos
	struct ObjectState 
	{
		ObjectState() : visible(true), meshIdx(0), pos(0,0,0), scale(1) {}
		ObjectState(bool isVisible, unsigned int meshidx, const Vector3D &position, float scaleF) : visible(isVisible), meshIdx(meshidx), pos(position), scale(scaleF) {}

		bool			visible;
		unsigned int	meshIdx;
		Vector3D		pos;
		float			scale;
	};
	typedef std::vector<ObjectState> ObjectStateVec;

	struct EyeState
	{
		EyeState(const Vector3D &p, const Vector3D &c, Vector3D &u) : pos(p), center(c), up(u) {}
		EyeState() : pos(0,0,0), center(0,0,0), up(0,0,0) {}

		Vector3D	pos;		//!< look from this position
		Vector3D	center;		//!< look at this position
		Vector3D	up;			//!< camera up vector
	};
	typedef std::list<EyeState> FlightPath;

    Q_OBJECT

public:            

    GLwidget(QWidget *parent = 0, QGLFormat format = QGLFormat());
	~GLwidget();
    
	bool isFlightPathAvailable() const { return (m_flightPath.size() > 0); }
	bool isVFCullingEnabled() const { return m_enableVFC; }

public slots:
	bool slotLoadModels(const QString &filename);
	void slotLoadConfiguration(const QString&);
	void slotSaveConfiguration(const QString&);
	void slotEnableVFC(bool b) { m_enableVFC = b; updateGL(); }
	void slotGenerateFlightPath();
	void slotPlayFlightPath();
	void slotLoadFlightPath(const QString&);
	void slotSaveFlightPath(const QString&);
	void slotNextPathPos();

signals:
	void signalRenderedObject(int);
	void signalTotalObjects(int);
	void signalFlightPathAvailable(bool);

protected:   
    virtual void paintGL();
	virtual void initializeGL();
	virtual void resizeGL ( int width, int height );
	virtual void mousePressEvent( QMouseEvent *event );
	virtual void mouseMoveEvent( QMouseEvent *event );

	float getRandomNumber( float range_min, float range_max) { return (double)rand() / (RAND_MAX + 1) * (range_max - range_min) + range_min; }
	void resetRotation(); // set rotation matrix to Identity
	
	void cleanUpScene();
	void cleanUpFlightPath();

	void computeFlightPath();								//!< INPUT: m_controlPoints, OUTPUT: m_flightPath
	float evalQuadraticBspline(float P0, float P1, float P2, float t);		//!< uniform quadratic B-spline
	Vector3D evalQuadraticBspline(Vector3D P0, Vector3D P1, Vector3D P2, float t);		//!< uniform quadratic B-spline

	void computeObjectVisibility();	//!< ** TODO **

	std::vector<Mesh*>			m_meshes;
	ObjectStateVec				m_meshStates;

	float						m_workVolumeSize;							//!< edge length of cubic rendering volume (object center is inside cube)
	GLdouble					m_left, m_right, m_bottom, m_top, m_Near, m_Far;	//!< View Frustum as set by glFrustum in resizeGL()

	FlightPath					m_controlPoints;							//!< control points of flight plan
	FlightPath					m_flightPath;								//!< interpolate controlPoints, evaluate at equidistant step size
	bool						m_enableVFC;
	float						m_flightStepSize;								//!< \in (0,1)
	bool						m_enableFlight;

	FlightPath::const_iterator	m_fpCurPos;									//!< current position in flight pas
	QTimer						*m_timer;		

	int							m_renderedObjects;							//!< sum of all rendered objects along path
	int							m_totalObjects;								//!< #objects * #frames at pathfly
	// some variables for a trackball
	float						m_RotationMatrix[16];						//!< global rotation Matrix
	float						m_xTrans, m_yTrans, m_zTrans;				//!< global translation
	QPoint						m_lastPos;

};


#endif
